#ifndef __ANIMATION_H_INCLUDED__
#define __ANIMATION_H_INCLUDED__


#include "geommath.h"
#include <string.h>

struct Key
{
	// user input
	float	frame;				// Key frame.

	// hermite based interpolation
	Vec3	Vec;					// 3 component vector used for hermite splines

	// delta rotation
	Quat	angleaxis;	 	// Quaternion in angle/axis.

	float	tens;					// Key tension value.
	float	cont;					// Key continuity value.
	float	bias;					// Key bias value.

	float	easeto;				// Key ease to value.
	float	easefrom;			// Key ease from value.

	// working variables
	Quat	A,B;				// Key derivatives.
	Quat	quatfromaxis;			// the quaternion representation of delta axisangle

	Vec3	tanin, tanout;
};


class IAnimCtrl
{
public:
	IAnimCtrl() {}
	~IAnimCtrl()  {}
	
	inline void		SetAnimInterval(float fFrameStart, float fFrameEnd)
	{
		m_fFrameStart = fFrameStart;
		m_fFrameEnd   = fFrameEnd;
	}

	inline void		SetAnimKeys(Key* pKeys, int iNumKeys)
	{
		m_pKeys		= pKeys;
		m_iNumKeys	= iNumKeys;
	}

	inline void		SetKey(int iKey, float frame,float x,float y,float z,float t,float c,float b,float eto,float efrom )
	{
		Key	&key	= m_pKeys[iKey];

		memset(&key, 0, sizeof(Key));
		key.frame	= frame;
		key.Vec		= Vec3(x,y,z);
		key.tens	= t;
		key.cont	= b;
		key.bias	= b;
		key.easeto	= eto;
		key.easefrom= efrom;
	}

	inline	void	SetRotKey(int iKey,	float frame,float x,float y,float z,float w,float t,float c,float b,float eto,float efrom )
	{
		Key	&key		= m_pKeys[iKey];
		
		memset(&key, 0, sizeof(Key));
		key.frame		= frame;
		key.angleaxis.V	= Vec3(x, y, z);
		key.angleaxis.s	= w;
		key.tens		= t;
		key.cont		= b;
		key.bias		= b;
		key.easeto		= eto;
		key.easefrom	= efrom;
	}

	inline	int		GetFrameStart()				{ return (int)m_fFrameStart; }
	inline	int		GetFrameEnd()				{ return (int)m_fFrameEnd; }
	inline	int		GetNumKeys()				{ return m_iNumKeys; }
	inline	Key		*GetKeys()					{ return m_pKeys; }
	inline	Key		&GetKey(int i)				{ return m_pKeys[i]; }



			virtual void	UpdateDerivates()=0;
			virtual void	UpdateQuatDerivates()=0;
			virtual void	EvalPos(float time, Vec3* vPos)=0;
			virtual void	EvalRot(float time, Quat* qRot)=0;
protected:
	
	int		m_iNumKeys;

	float	m_fFrameStart;
	float	m_fFrameEnd;

	Key		*m_pKeys;

private:
};

#endif	//__ANIMATION_H_INCLUDED__