#ifndef __CONV3DS_H__
#define __CONV3DS_H__

#include "internscene.h"
#include <string>

class C3DModel;
class C3DObject;
class C3DFace;
class C3DCamera;
class C3DLight;
class CTexturePropContainer;
class CConvTempAnim;

class CConv3DS : public CInternScene
{
public:
	bool Conv3dsToIntern(const char * pFile, CTexturePropContainer * pTexPropContainer);
	CConv3DS() : m_bForceYaxisUp(true), m_pTmpAnim(NULL), m_fFPS(30.0f) {}

private:
	const Vec3 ConvMeshToIntern(CInternMesh * pIntMesh, C3DObject * p3dsObj);
	void MakeInternMeshInstance(CInternMeshInstance * pMeshInst, CInternMesh * pMesh, const string name, C3DObject * p3dsObj, const Vec3 vCen, bool bFoundAnim);
	void MakeAllMeshInstances(C3DModel &c3dsScene, Vec3 * centers);
	void MakeAllDummyInstances(C3DModel &c3dsScene);
	
	int DetermineNumSurfaces(C3DObject * p3dsObj);	
	string FetchSurfName(int32 iSurfNum, C3DObject * p3dsObj);
	bool isFaceValid(C3DFace &face, C3DObject * p3dsObj);

private:
	// Camera creation
	void MakeInternCamera(CInternCamera * pIntCam, const string name, const C3DCamera &c3dsCam, bool bFoundAnim);
	void MakeAllCameras(C3DModel &c3dsScene);

	// Light creation
	void MakeInternLight(CInternLight * pIntLgt, const string name, const C3DLight &c3dsLgt, bool bFoundAnim);
	void MakeAllLights(C3DModel &c3dsScene);

private:
	int32 DetermineNumMaterials(C3DModel &c3dmod);
	void DistributeMatNames(C3DModel &c3dmod);
	void DistributeMatToSurf(C3DModel &c3dmod);
	void DistributeTextureSequences(const C3DModel &c3dsScene, CTexturePropContainer * pTexPropContainer, const char input_file[]);
	void FetchNameOfScene(const char * pFile);

private:	// geometry builders
	int32 CreateFaceLists(CInternMesh * pIntMesh, C3DObject * p3dsObj, int32 *** pdescs);
	void CreateNormals(C3DObject * p3dsObj, Vec3 * pFinalNormals, SInternFace * trngNrms);
	void CreateTangBinVectors(C3DObject * p3dsObj, const Vec3 vNormals[], const SInternFace trngNorms[], Vec3 * vS, Vec3 * vT, SInternFace * trngTang);
	void CreateGeometryData(CInternMesh * pIntMesh, C3DObject * p3dsObj, const Vec3 * pFinalNorms, const SInternFace * trngNrms, const Vec3 * pTangVectors, const Vec3 * pBinoVectors, const SInternFace * trngTang, const int32 ** descs, const int32 max_trng);

private: // spline fetchers
	const int32 Max(const int32 a, const int32 b);
	void CreateTransform(CTransformation * pTrans, const Vec3 &vCen, const Mat33 &mDefault33, const Vec3 &vDefault3, const bool bAnimFound);
	void CreateKey1Lst(CInternKey1List * pKey1Lst, const float fDefault, const bool bAnimFound);
	void CreateKey3Lst(CInternKey3List * pKey3Lst, const Vec3 &vDefault, const bool bAnimFound);


private:	// utility functions
	const Mat33 CalcRot(const Vec3 &vPos, const Vec3 &vTarget, const float roll);
	const Mat33 GetAxisUpMatrix();
	bool m_bForceYaxisUp;

private: // anim finder
	int32 GetMeshIndex(C3DModel &c3dsScene, string Name);
	int32 GetCameraIndex(C3DModel &c3dsScene, string Name);
	int32 GetOmniIndex(C3DModel &c3dsScene, string Name);
	int32 GetSpotIndex(C3DModel &c3dsScene, string Name);

	CConvTempAnim * m_pTmpAnim;

private:	// animtime
	float m_fFPS;	// Frames played per second. (default is ntsc, 30)
};

#endif


