#ifndef __CONVTEMPANIM_H__
#define __CONVTEMPANIM_H__

#include "exttypes.h"
#include "internspline.h"
#include <string>


class C3DModel;
class C3DTrackInfo;
class C3DTrackCamera;
class C3DTrackOmni;
class C3DTrackSpot;
class C3DTrackCamTarget;
class C3DTrackSpotTarget;
class IAnimCtrl;

class CConvTempAnim
{
public:	
	void Setup(const C3DModel &c3dsScene);

	// Indexed mesh related data and settings
	const int32 GetNumMeshAnims() const { return m_iNumMeshAnims; }
	const int32 GetNumCameraAnims() const { return m_iNumCamAnims; }
	const int32 GetNumOmniAnims() const { return m_iNumOmniAnims; }
	const int32 GetNumSpotAnims() const { return m_iNumSpotAnims; }

	bool SetMesh(const int32 meshindex) { m_pTmpCurTrans=&m_pTmpMeshAnim[meshindex]; node_mode=MESH_NODE; return m_pTmpCurTrans!=NULL; }
	bool SetCamera(const int32 cameraindex) { m_pTmpCurTrans=&m_pTmpCamAnim[cameraindex]; node_mode=CAMERA_NODE; return m_pTmpCurTrans!=NULL; }
	bool SetOmni(const int32 omniindex) { m_pTmpCurTrans=&m_pTmpOmniAnim[omniindex]; node_mode=OMNI_NODE; return m_pTmpCurTrans!=NULL; }
	bool SetSpot(const int32 spotindex) { m_pTmpCurTrans=&m_pTmpSpotAnim[spotindex]; node_mode=SPOT_NODE; return m_pTmpCurTrans!=NULL; }

	string GetName() const;
	string GetTrackInstName() const;

		
	// Mesh related data and settings
	bool SetMesh(const string meshName) { m_pTmpCurTrans=(CTempMeshAnim *) LookUpAnim(meshName, m_pTmpMeshAnim, m_iNumMeshAnims); node_mode=MESH_NODE; return m_pTmpCurTrans!=NULL; }
	bool SetCamera(const string camName) { m_pTmpCurTrans=(CTempCamAnim *) LookUpAnim(camName, m_pTmpCamAnim, m_iNumCamAnims); node_mode=CAMERA_NODE; return m_pTmpCurTrans!=NULL; }
	bool SetOmni(const string omniName) { m_pTmpCurTrans=(CTempOmniAnim *) LookUpAnim(omniName, m_pTmpOmniAnim, m_iNumOmniAnims); node_mode=OMNI_NODE; return m_pTmpCurTrans!=NULL; }
	bool SetSpot(const string spotName) { m_pTmpCurTrans=(CTempSpotAnim *) LookUpAnim(spotName, m_pTmpSpotAnim, m_iNumSpotAnims); node_mode=SPOT_NODE; return m_pTmpCurTrans!=NULL; }

	void SetFovKeyList() { attrib_type=FOV_GET; }
	void SetHotSpotKeyList() { attrib_type=HOTSPOT_GET; }
	void SetFallOffKeyList() { attrib_type=FALLOFF_GET; }
	void SetPositionKeyList() { attrib_type=POSITION_GET; }
	void SetRotationKeyList() { attrib_type=ROTATION_GET; }
	void SetScaleKeyList() { attrib_type=SCALE_GET; }

	const int32 GetNumKey1s() const;
	const int32 GetNumKey3s() const;
	const int32 GetNumMat33s() const;

	const float GetKey1(const int32 index) const;
	const Vec3 GetKey3(const int32 index) const;
	const Mat33 GetMat33(const int32 index) const;


	CConvTempAnim();

	
private:
	class CTempMat33List
	{
		friend class CConvTempAnim;

	public:
		const int32 GetNumKeys() const { return m_iNumKeys; }
		const Mat33 GetKey(const int32 index) const { return m_pKeys[index]; }
		CTempMat33List() : m_iNumKeys(0), m_pKeys(NULL) {}

	private:
		int32 m_iNumKeys;
		Mat33 * m_pKeys;
	};

	
	class CTempTransform
	{
	public:
		CInternKey3List m_cTranslation;
		CTempMat33List m_cRotation;
		CInternKey1List m_cUniScale;

		int32 m_iHierarchyNum;
		int32 m_iNodeID;
		string m_s3dsName;
		string m_s3dsTrackInstName;

		CTempTransform() : m_pParent(NULL), m_pNextSibling(NULL), m_pFirstChild(NULL) {}

	public:
		CTempTransform * m_pParent;
		CTempTransform * m_pFirstChild, * m_pNextSibling;
	};

	class CTempMeshAnim : public CTempTransform
	{
	public:
		
	};

	class CTempCamAnim : public CTempTransform
	{
	public:
		CInternKey1List m_cRoll;
		CInternKey1List m_cFov;
	};

	class CTempOmniAnim : public CTempTransform
	{
	public:
		
	};

	class CTempSpotAnim : public CTempTransform
	{
	public:
		CInternKey1List m_cRoll;
		CInternKey1List m_cHotSpot;
		CInternKey1List m_cFallOff;
	};

	class CTempCamTgtAnim : public CTempTransform
	{
	public:
		
	};

	class CTempSpotTgtAnim : public CTempTransform
	{
	public:
		
	};

	int32 m_iNumMeshAnims;
	int32 m_iNumCamAnims;
	int32 m_iNumOmniAnims;
	int32 m_iNumSpotAnims;
	int32 m_iNumCamTargetAnims;
	int32 m_iNumSpotTargetAnims;

	
	CTempMeshAnim * m_pTmpMeshAnim;
	CTempCamAnim * m_pTmpCamAnim;
	CTempOmniAnim * m_pTmpOmniAnim;
	CTempSpotAnim * m_pTmpSpotAnim;
	CTempCamTgtAnim * m_pTmpCamTargetAnim;
	CTempSpotTgtAnim * m_pTmpSpotTargetAnim;

private:
	void Do3dsToTempMeshAnim(CTempMeshAnim * pMeshAnim, const C3DTrackInfo * pTrackInfo);
	void Do3dsToTempCamAnim(CTempCamAnim * pCamAnim, const C3DTrackCamera * pTrackCam);
	void Do3dsToTempOmniAnim(CTempOmniAnim * pOmniAnim, const C3DTrackOmni * pTrackOmni);
	void Do3dsToTempSpotAnim(CTempSpotAnim * pSpotAnim, const C3DTrackSpot * pTrackSpot);
	void Do3dsToTempCamTargetAnim(CTempCamTgtAnim * pCamTargetAnim, const C3DTrackCamTarget * pTrackCamTarget);
	void Do3dsToTempSpotTargetAnim(CTempSpotTgtAnim * pSpotTargetAnim, const C3DTrackSpotTarget * pTrackSpotTarget);

	void Do3dsToTempTransAnim(CTempTransform * pTransOut, const C3DTrackInfo * pTransIn);

private:
	const Vec3 GetPosition(const float t, const int num_keys);
	const Mat33 GetRotation(const float t, const int num_keys);
	const float GetUniScale(const float t, const int num_keys);
	
	int32 m_iFrameBegin;
	int32 m_iFrameEnd;

	IAnimCtrl * m_pAnimCtrl;

private:
	void HandleHierarchy(CTempTransform ** ppTmpTrans, const int32 iTotNumNodesInHierarchy);
	void ApplyRecurTransform(CTempTransform * pCurNode);

private:
	void FixupTransDirectional(CTempTransform * pTmpTrans, const CInternKey1List &cRoll, const CTempTransform m_TmpTargetAnim[], const int32 num_targets);
	const Mat33 CalcRot(const Vec3 &vPos, const Vec3 &vTarget, const float roll);

private:
	void DownSampleMeshSplines(CTempMeshAnim * pMeshAnim);
	void DownSampleCameraSplines(CTempCamAnim * pCamAnim);
	void DownSampleOmniSplines(CTempOmniAnim * pOmniAnim);
	void DownSampleSpotSplines(CTempSpotAnim * pSpotAnim);
	void DeleteCamTargetSplines(CTempCamTgtAnim * pCamTargetAnim);
	void DeleteSpotTargetSplines(CTempSpotTgtAnim * pSpotTargetAnim);

	void DownSampleTempTrans(CTempTransform * pTransAnim);
	void DownSampleKey1Lst(CInternKey1List * pKeyLst);
	void DownSampleKey3Lst(CInternKey3List * pKeyLst);
	void DownSampleTempMat33Lst(CTempMat33List * pKeyLst);

	void CloseDownTempTrans(CTempTransform * pTransAnim);
	void CloseDownKey1Lst(CInternKey1List * pKeyLst);
	void CloseDownKey3Lst(CInternKey3List * pKeyLst);
	void CloseDownTempMat33Lst(CTempMat33List * pKeyLst);

private:
	template <class _T> CTempTransform * LookUpAnim(string name, _T search_pile[], const int32 num_in_pile);
	
	enum NODE_SET
	{
		MESH_NODE=1,
		CAMERA_NODE,
		OMNI_NODE,
		SPOT_NODE,

		NODE_UNKNOWN
	};
	NODE_SET node_mode;

	enum ATTRIB_GET
	{
		FOV_GET=1,
		HOTSPOT_GET,
		FALLOFF_GET,
		POSITION_GET,
		ROTATION_GET,
		SCALE_GET,

		ATTRIB_UNKNOWN
	};
	ATTRIB_GET attrib_type;
	
	CTempTransform * m_pTmpCurTrans;
};

#endif
