#include "export.h"
#include "internscene.h"
#include "interntextureseq.h"
#include "pixel_formats.h"
#include <limits.h>
#include <stdio.h>
#include "tc.h"



void CExport::WriteScene(CInternScene &sceneIn, const char outFolder[], const int32 strip_maxverts)
{
	// create path+names
	int len_outfolder = strlen(outFolder);
	int len_scenename = strlen(sceneIn.GetName());
	int len_sceneprefix = strlen("scene_");
	int len_binprefix = strlen("geomdata_");

	int len_scenepathname = len_outfolder + len_sceneprefix + len_scenename + strlen(".asc") + 1;	// '\0'
	int len_binpathname = len_outfolder + len_binprefix + len_scenename + strlen(".bin") + 1;	// '\0'

	// need to seperate path from names
	bool extend_by_slash = false;
	if(len_outfolder>0 && outFolder[len_outfolder-1]!='/' && outFolder[len_outfolder-1]!='\\' && outFolder[len_outfolder-1]!=':')
	{
		++len_scenepathname;
		++len_binpathname;
		extend_by_slash = true;
	}

	char * pOutScene = NULL;
	char * pOutBin = NULL;
	pOutScene = new char[len_scenepathname];
	pOutBin = new char[len_binpathname];
	assert(pOutScene!=NULL && pOutBin!=NULL);

	strncpy(pOutScene, outFolder, len_outfolder);
	strncpy(pOutBin, outFolder, len_outfolder);

	char * cptr1 = pOutScene + len_outfolder;
	char * cptr2 = pOutBin + len_outfolder;
	if(extend_by_slash)
	{
		*cptr1++ = '/';
		*cptr2++ = '/';
	}

	strncpy(cptr1, "scene_", len_sceneprefix);
	strncpy(cptr2, "geomdata_", len_binprefix);
	cptr1 += len_sceneprefix;
	cptr2 += len_binprefix;

	strncpy(cptr1, sceneIn.GetName(), len_scenename);
	strncpy(cptr2, sceneIn.GetName(), len_scenename);
	cptr1 += len_scenename;
	cptr2 += len_scenename;
	strcpy(cptr1, ".asc");
	strcpy(cptr2, ".bin");
	assert(cptr1[4]=='\0' && cptr2[4]=='\0');

	
	// Get max tris in any surface
	int32 iNumSceSurfs = 0, m;
	m_iMaxTriCount = 0;
	for(m=0; m<sceneIn.GetNumMeshes(); m++)
	{
		CInternMesh &mesh = sceneIn.GetMesh(m);

		// verify this is a unique name
		for(int m2=0; m2<m; m2++)
			assert(strcmp(sceneIn.GetMesh(m2).GetName(), mesh.GetName()) != 0);
		
		iNumSceSurfs += mesh.GetNumSurfaces();
		
		for(int s=0; s<mesh.GetNumSurfaces(); s++)
		{
			CInternSurface &surf = mesh.GetSurface(s);
			if(m_iMaxTriCount < surf.GetNumTriangles())
				m_iMaxTriCount = surf.GetNumTriangles();
		}
	}
	
	// Determine number of unique texture frames used by this scene
	if(m_pUniqueTextures != NULL) { delete [] m_pUniqueTextures; m_pUniqueTextures=NULL; }
	m_iNumUniqueTextures=0;
	for(m=0; m<sceneIn.GetNumMaterials(); m++)
	{
		AddInternTextureSeq(sceneIn.GetMaterial(m).GetDiffSequence());
		AddInternTextureSeq(sceneIn.GetMaterial(m).GetSpecSequence());
		AddInternTextureSeq(sceneIn.GetMaterial(m).GetReflSequence());
		AddInternTextureSeq(sceneIn.GetMaterial(m).GetBumpSequence());
	}

	// Determine number of splines with more entries then 1
	m_iNumSplines = 0;
	for(int32 m=0; m<sceneIn.GetNumMeshInstances(); m++)
	{
		if(sceneIn.GetMeshInstance(m).GetNumPosKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetMeshInstance(m).GetNumRotKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetMeshInstance(m).GetNumScaleKeys()>1)
			++m_iNumSplines;
	}
	for(int32 c=0; c<sceneIn.GetNumCameras(); c++)
	{
		if(sceneIn.GetCamera(c).GetNumPosKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetCamera(c).GetNumRotKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetCamera(c).GetNumScaleKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetCamera(c).GetNumFovKeys()>1)
			++m_iNumSplines;
	}
	for(int32 l=0; l<sceneIn.GetNumLights(); l++)
	{
		if(sceneIn.GetLight(l).GetNumPosKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetLight(l).GetNumRotKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetLight(l).GetNumScaleKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetLight(l).GetNumColorKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetLight(l).GetNumNearKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetLight(l).GetNumFarKeys()>1)
			++m_iNumSplines;
		if(sceneIn.GetLight(l).IsSpot()==true)
		{
			if(sceneIn.GetLight(l).GetNumHotSpotKeys()>1)
				++m_iNumSplines;
			if(sceneIn.GetLight(l).GetNumFallOffKeys()>1)
				++m_iNumSplines;
			if(sceneIn.GetLight(l).GetNumAspectKeys()>1)
				++m_iNumSplines;
		}
	}

	
	// alloc tmp ID buffer
	m_iTotNumSurfs = 0;
	for(m=0; m<sceneIn.GetNumMeshes(); m++)
		for(int s=0; s<(sceneIn.GetMesh(m)).GetNumSurfaces(); s++)
			++m_iTotNumSurfs;
	m_iTotIDs = m_iTotNumSurfs + m_iNumUniqueTextures + m_iNumSplines;
	if(m_pTotIDs != NULL) delete [] m_pTotIDs;
	m_pTotIDs = new int32[m_iTotIDs];	// for ASCII file

	// Write Binary data	
	m_iIndexCounter = 0;	// to buffer of strip lengths for ASCII file
	if(m_fBinFile != NULL) fclose(m_fBinFile);
	m_fBinFile = NULL;
	if((m_fBinFile=fopen(pOutBin, "wb")) == NULL)
	{
		printf("failed to make output geomdata file: %s\n", pOutBin);
		exit(0);
	}
	WriteGeometryToBinFile(sceneIn, strip_maxverts);
	WriteTexturesToBinFile(sceneIn);
	WriteSplinesToBinFile(sceneIn);	
	fclose(m_fBinFile); m_fBinFile = NULL;
	

	// Write ASCII representation of the scene
	m_iIndexCounter = 0;
	if(m_fSceneFile != NULL) fclose(m_fSceneFile);
	m_fSceneFile = NULL;
	if((m_fSceneFile=fopen(pOutScene, "wb")) == NULL)
	{
		printf("failed to make output scene file: %s\n", pOutScene);
		exit(0);
	}
	WriteSceneToAscFile(sceneIn);
	
	delete [] m_pTotIDs;
	delete [] pOutScene; delete [] pOutBin;
	fclose(m_fSceneFile); m_fSceneFile = NULL;
}

//////////////////////////////////////////////////////////////////////////////////
	

// Write binary data to file
void CExport::WriteGeometryToBinFile(CInternScene &sceneIn, const int32 strip_maxverts)
{
	// allocations for strip finder
	unsigned int * triangles = new unsigned int[3*m_iMaxTriCount];
	int * primLengths = new int[m_iMaxTriCount];
	unsigned int * primVerts = new unsigned int[3*m_iMaxTriCount];
	int * primTypes = new int[m_iMaxTriCount];

	// Write binary representations of all strips
	for(int32 m=0; m<sceneIn.GetNumMeshes(); m++)
	{
		CInternMesh &mesh = sceneIn.GetMesh(m);
		
		for(int s=0; s<mesh.GetNumSurfaces(); s++)
		{
			CInternSurface &surf = mesh.GetSurface(s);

			// Start a new chunk entry in the binary file
			PushBinChunk();

			// locate strips
			int32 iNumStrips = GenStrips(surf, triangles, primTypes, primLengths, primVerts, strip_maxverts);	

			// output length of vertex arrays
			int32 tmp = surf.GetArrayLen();
			fwrite(&tmp, sizeof(int32), 1, m_fBinFile);

			// output number of strips
			fwrite(&iNumStrips, sizeof(int32), 1, m_fBinFile);

			// output vertices
			for(int32 v=0; v<surf.GetArrayLen(); v++)
			{
				Vec3 &vec = surf.GetVertex(v);
				fwrite(&vec.x, sizeof(float), 1, m_fBinFile);
				fwrite(&vec.y, sizeof(float), 1, m_fBinFile);
				fwrite(&vec.z, sizeof(float), 1, m_fBinFile);
			}

			// output normals
			for(int32 n=0; n<surf.GetArrayLen(); n++)
			{
				Vec3 &norm = surf.GetNormal(n);
				fwrite(&norm.x, sizeof(float), 1, m_fBinFile);
				fwrite(&norm.y, sizeof(float), 1, m_fBinFile);
				fwrite(&norm.z, sizeof(float), 1, m_fBinFile);
			}

			// output tangentvectors
			for(int32 t=0; t<surf.GetArrayLen(); t++)
			{
				Vec3 &tang = surf.GetTangent(t);
				fwrite(&tang.x, sizeof(float), 1, m_fBinFile);
				fwrite(&tang.y, sizeof(float), 1, m_fBinFile);
				fwrite(&tang.z, sizeof(float), 1, m_fBinFile);
			}

			// output binormals
			for(int32 b=0; b<surf.GetArrayLen(); b++)
			{
				Vec3 &bino = surf.GetBiNorm(b);
				fwrite(&bino.x, sizeof(float), 1, m_fBinFile);
				fwrite(&bino.y, sizeof(float), 1, m_fBinFile);
				fwrite(&bino.z, sizeof(float), 1, m_fBinFile);
			}

			// output STs
			for(int32 t=0; t<surf.GetArrayLen(); t++)
			{
				Vec3 &st = surf.GetST(t);
				fwrite(&st.x, sizeof(float), 1, m_fBinFile);
				fwrite(&st.y, sizeof(float), 1, m_fBinFile);
			}
			
			// output strips
			unsigned int * pTmps = primVerts;
			for(int k=0; k<iNumStrips; k++)
			{
				int32 iLen = (int32) primLengths[k];
				fwrite(&iLen, sizeof(int32), 1, m_fBinFile);
				for(int i=0; i<iLen; i++)
				{ int32 iIndex=(int32) *pTmps++; fwrite(&iIndex, sizeof(int32), 1, m_fBinFile); }
			}

			// Mark this chunk as done
			PopBinChunk();
		}
	}
	delete [] triangles; delete [] primLengths;
	delete [] primVerts; delete [] primTypes;
	assert(m_iIndexCounter == m_iTotNumSurfs);
}

void CExport::WriteTexturesToBinFile(CInternScene &sceneIn)
{
	// Write binary representations of all unique textures
	for(int t=0; t<m_iNumUniqueTextures; t++)
	{
		// Start a new chunk entry in the binary file
		PushBinChunk();

		// next texture
		const CInternTexture * pTex = m_pUniqueTextures[t];

		// pix_size
		int32 pix_size;
		switch(pTex->GetPixelFormat())
		{
			case INDEXCOLOR16:
				pix_size = 1;
				break;
			case INDEXCOLOR256:
				pix_size = 1;
				break;
			case COLORRGBA32:
				pix_size = 4;
				break;
			default:
				assert(false);	// unknown format
		};

		// Now write data into file
		if(pTex->GetUserDefProps()!=NULL)
			fwrite(pTex->GetUserDefProps(), sizeof(char), strlen(pTex->GetUserDefProps()) + 1, m_fBinFile);
		else
			fputc('\0', m_fBinFile);

		// write format of pixels
		const int32 pix_format = pTex->GetPixelFormat();
		fwrite(&pix_format, sizeof(uint32), 1, m_fBinFile);

		// write palette of there is one
		if(pTex->GetPixelFormat()==INDEXCOLOR16)
			fwrite(pTex->GetPalette(), sizeof(uint8), 16*4, m_fBinFile);
		else if(pTex->GetPixelFormat()==INDEXCOLOR256)
			fwrite(pTex->GetPalette(), sizeof(uint8), 256*4, m_fBinFile);

		// write num levels
		const int32 num_levels = pTex->GetNumLevels();
		fwrite(&num_levels, sizeof(uint32), 1, m_fBinFile);

		for(int l=0; l<pTex->GetNumLevels(); l++)
		{
			// write width and height into file
			const int32 width = pTex->GetLevel(l)->GetWidth();
			const int32 height = pTex->GetLevel(l)->GetHeight();
			fwrite(&width, sizeof(uint32), 1, m_fBinFile);
			fwrite(&height, sizeof(uint32), 1, m_fBinFile);

			// write pix_data into file
			fwrite(pTex->GetLevel(l)->GetPixData(), sizeof(uint8), width*height*pix_size, m_fBinFile);
		}

		// Mark this chunk as done
		PopBinChunk();
	}

	assert(m_iIndexCounter == m_iNumUniqueTextures+m_iTotNumSurfs);
}

void CExport::WriteSplinesToBinFile(CInternScene &sceneIn)
{
	// Write binary representations of splines
	for(int32 m=0; m<sceneIn.GetNumMeshInstances(); m++)
	{
		if(sceneIn.GetMeshInstance(m).GetNumPosKeys()>1)
			WriteBinKey3List(sceneIn.GetMeshInstance(m).m_PosArray);
		if(sceneIn.GetMeshInstance(m).GetNumRotKeys()>1)
			WriteBinQuatList(sceneIn.GetMeshInstance(m).m_RotArray);
		if(sceneIn.GetMeshInstance(m).GetNumScaleKeys()>1)
			WriteBinKey1List(sceneIn.GetMeshInstance(m).m_ScaleArray);
	}
	for(int32 c=0; c<sceneIn.GetNumCameras(); c++)
	{
		if(sceneIn.GetCamera(c).GetNumFovKeys()>1)
			WriteBinKey1List(sceneIn.GetCamera(c).m_FovArray);
		if(sceneIn.GetCamera(c).GetNumPosKeys()>1)
			WriteBinKey3List(sceneIn.GetCamera(c).m_PosArray);
		if(sceneIn.GetCamera(c).GetNumRotKeys()>1)
			WriteBinQuatList(sceneIn.GetCamera(c).m_RotArray);
		if(sceneIn.GetCamera(c).GetNumScaleKeys()>1)
			WriteBinKey1List(sceneIn.GetCamera(c).m_ScaleArray);
	}
	for(int32 l=0; l<sceneIn.GetNumLights(); l++)
	{
		if(sceneIn.GetLight(l).GetNumColorKeys()>1)
			WriteBinKey3List(sceneIn.GetLight(l).m_ColorArray);
		if(sceneIn.GetLight(l).GetNumNearKeys()>1)
			WriteBinKey1List(sceneIn.GetLight(l).m_NearArray);
		if(sceneIn.GetLight(l).GetNumFarKeys()>1)
			WriteBinKey1List(sceneIn.GetLight(l).m_FarArray);

		if(sceneIn.GetLight(l).IsSpot()==true)
		{
			if(sceneIn.GetLight(l).GetNumHotSpotKeys()>1)
				WriteBinKey1List(sceneIn.GetLight(l).m_HotSpotArray);
			if(sceneIn.GetLight(l).GetNumFallOffKeys()>1)
				WriteBinKey1List(sceneIn.GetLight(l).m_FallOffArray);
			if(sceneIn.GetLight(l).GetNumAspectKeys()>1)
				WriteBinKey1List(sceneIn.GetLight(l).m_AspectArray);
		}

		if(sceneIn.GetLight(l).GetNumPosKeys()>1)
			WriteBinKey3List(sceneIn.GetLight(l).m_PosArray);
		if(sceneIn.GetLight(l).GetNumRotKeys()>1)
			WriteBinQuatList(sceneIn.GetLight(l).m_RotArray);
		if(sceneIn.GetLight(l).GetNumScaleKeys()>1)
			WriteBinKey1List(sceneIn.GetLight(l).m_ScaleArray);
	}

	assert(m_iIndexCounter == m_iNumUniqueTextures+m_iTotNumSurfs+m_iNumSplines);
}


void CExport::WriteBinKey1List(const CInternKey1List &keyList)
{
	// Start a new chunk entry in the binary file
	PushBinChunk();

	// write data
	int32 num_keys = keyList.GetNumKeys();
	int32 num_channels = 1;

	fwrite(&num_keys, sizeof(int32), 1, m_fBinFile);
	fwrite(&num_channels, sizeof(int32), 1, m_fBinFile);

	for(int32 k=0; k<keyList.GetNumKeys(); k++)
	{
		float val = keyList.GetKey(k);
		fwrite(&val, sizeof(float), 1, m_fBinFile);
	}

	// Mark this chunk as done
	PopBinChunk();
}

void CExport::WriteBinKey3List(const CInternKey3List &keyList)
{
	// Start a new chunk entry in the binary file
	PushBinChunk();

	// write data
	int32 num_keys = keyList.GetNumKeys();
	int32 num_channels = 3;

	fwrite(&num_keys, sizeof(int32), 1, m_fBinFile);
	fwrite(&num_channels, sizeof(int32), 1, m_fBinFile);

	for(int32 k=0; k<keyList.GetNumKeys(); k++)
	{
		const Vec3 v = keyList.GetKey(k);
		float x=v.x, y=v.y, z=v.z;
		fwrite(&x, sizeof(float), 1, m_fBinFile);
		fwrite(&y, sizeof(float), 1, m_fBinFile);
		fwrite(&z, sizeof(float), 1, m_fBinFile);
	}

	// Mark this chunk as done
	PopBinChunk();
}

void CExport::WriteBinQuatList(const CInternQuatList &keyList)
{
	// Start a new chunk entry in the binary file
	PushBinChunk();

	// write data
	int32 num_keys = keyList.GetNumKeys();
	int32 num_channels = 4;

	fwrite(&num_keys, sizeof(int32), 1, m_fBinFile);
	fwrite(&num_channels, sizeof(int32), 1, m_fBinFile);

	for(int32 k=0; k<keyList.GetNumKeys(); k++)
	{
		const Quat q = keyList.GetKey(k);
		float x=q.V.x, y=q.V.y, z=q.V.z, w=q.s;
		fwrite(&x, sizeof(float), 1, m_fBinFile);
		fwrite(&y, sizeof(float), 1, m_fBinFile);
		fwrite(&z, sizeof(float), 1, m_fBinFile);
		fwrite(&w, sizeof(float), 1, m_fBinFile);
	}

	// Mark this chunk as done
	PopBinChunk();
}


//////////////////////////////////////////////////////////////////////////////////


void CExport::WriteSceneToAscFile(CInternScene &sceneIn)
{
	// write anim time
	fprintf(m_fSceneFile, "animtime<%f>\t// in seconds\n\n\n", sceneIn.GetAnimTime());
	
	// write materials
	for(int32 m=0; m<sceneIn.GetNumMaterials(); m++)
	{
		CInternMaterial &material = sceneIn.GetMaterial(m);
		
		// verify this is a unique name
		for(int m2=0; m2<m; m2++)
			assert(strcmp(sceneIn.GetMaterial(m2).GetName(),material.GetName()) != 0);
		
		const char * matName = material.GetName();

		fprintf(m_fSceneFile, "material<\"%s\">\n{\n", matName);

		// write diffuse texture
		WriteTexSeqToSceneFile(m_fSceneFile, material.GetDiffSequence(), "diffuse", m_pTotIDs+m_iTotNumSurfs);
		WriteTexSeqToSceneFile(m_fSceneFile, material.GetSpecSequence(), "specular", m_pTotIDs+m_iTotNumSurfs);
		WriteTexSeqToSceneFile(m_fSceneFile, material.GetReflSequence(), "reflection", m_pTotIDs+m_iTotNumSurfs);
		WriteTexSeqToSceneFile(m_fSceneFile, material.GetBumpSequence(), "bumpmap", m_pTotIDs+m_iTotNumSurfs);

		fprintf(m_fSceneFile, "}\n\n\n");
	}
	delete [] m_pUniqueTextures; m_pUniqueTextures=NULL; m_iNumUniqueTextures=0;
	
	// write meshes
	for(int32 m=0; m<sceneIn.GetNumMeshes(); m++)
	{
		CInternMesh &mesh = sceneIn.GetMesh(m);
		const char * objName = mesh.GetName();

		fprintf(m_fSceneFile, "mesh<\"%s\">\n{\n", objName);
		
		for(int s=0; s<mesh.GetNumSurfaces(); s++)
		{
			CInternSurface &surf = mesh.GetSurface(s);
			const char * matName = surf.GetMaterial()->GetName();
			char cID[ID_STRING_LENGTH];
			IDtoString(cID);	
			fprintf(m_fSceneFile, "\tsurface<\"%s\", \"%s\">\n", cID, matName);
		}
		
		fprintf(m_fSceneFile, "}\n\n\n");
	}
	assert(m_iIndexCounter == m_iTotNumSurfs);

	// write mesh instances
	for(int32 m=0; m<sceneIn.GetNumMeshInstances(); m++)
	{
		CInternMeshInstance &meshinst = sceneIn.GetMeshInstance(m);

		fprintf(m_fSceneFile, "meshinstance<\"%s\">\n{\n", meshinst.GetName());
		fprintf(m_fSceneFile, "\tmeshref<\"%s\">\n", meshinst.GetMesh().GetName());
			
		WriteAsciiTransformation( sceneIn.GetMeshInstance(m) );
		
		fprintf(m_fSceneFile, "}\n\n\n");
	}

	// write cameras
	for(int c=0; c<sceneIn.GetNumCameras(); c++)
	{
		CInternCamera &cam = sceneIn.GetCamera(c);
		
		fprintf(m_fSceneFile, "camera<\"%s\">\n{\n", cam.GetName());
		if(cam.GetNumFovKeys()>1)
		{
			char cID[ID_STRING_LENGTH];
			IDtoString(cID);
			fprintf(m_fSceneFile, "\tfieldofviewarray<\"%s\">\n", cID);
		}
		else	
			fprintf(m_fSceneFile, "\tfieldofview<%f>\n", cam.GetFovKey(0));

		fprintf(m_fSceneFile, "\tratio<[%d:%d]>\n", cam.GetRatioX(), cam.GetRatioY());

		WriteAsciiTransformation( cam );

		fprintf(m_fSceneFile, "}\n\n\n");
	}

	// write lights
	for(int l=0; l<sceneIn.GetNumLights(); l++)
	{
		CInternLight &light = sceneIn.GetLight(l);

		fprintf(m_fSceneFile, "light<\"%s\">\n{\n", light.GetName());
		WriteAsciiKey3Lst("color", light.m_ColorArray);
		WriteAsciiKey1Lst("near", light.m_NearArray);
		WriteAsciiKey1Lst("far", light.m_FarArray);

		// spot related settings
		if(light.IsSpot()==true)
		{	
			WriteAsciiKey1Lst("hotspot", light.m_HotSpotArray);
			WriteAsciiKey1Lst("falloff", light.m_FallOffArray);
			WriteAsciiKey1Lst("aspect", light.m_AspectArray);

			// output shape
			if(light.IsRectangular()==true)
				fprintf(m_fSceneFile, "\tspotshape<%s>\n", "RECTANGULAR");
			else
				fprintf(m_fSceneFile, "\tspotshape<%s>\n", "CIRCULAR");
		}

		// transformation
		WriteAsciiTransformation( light );
		
		fprintf(m_fSceneFile, "}\n\n\n");
	}
}

void CExport::WriteAsciiTransformation(const CTransformation &trans)
{	
	// deliver positions
	WriteAsciiKey3Lst("position", trans.m_PosArray);
		
	// deliver rotations
	char cID[ID_STRING_LENGTH];
	if(trans.GetNumRotKeys()>1)
	{
		IDtoString(cID);	
		fprintf(m_fSceneFile, "\tquaternionarray<\"%s\">\n", cID);
	}
	else
		fprintf(m_fSceneFile, "\tquaternion<%f, %f, %f, %f>\n", trans.GetRotKey(0).V.x, trans.GetRotKey(0).V.y, trans.GetRotKey(0).V.z, trans.GetRotKey(0).s);

	// deliver scale
	if(trans.GetNumScaleKeys()>1)
	{
		IDtoString(cID);	
		fprintf(m_fSceneFile, "\tscalearray<\"%s\">\n", cID);
	}
	else if(trans.GetScaleKey(0) != 1.0f)
		fprintf(m_fSceneFile, "\tscale<%f>\n", trans.GetScaleKey(0));
}


void CExport::WriteAsciiKey1Lst(const char cptrToken[], const CInternKey1List &key1lst)
{
	if(key1lst.GetNumKeys()>1)
	{
		char cID[ID_STRING_LENGTH];
		IDtoString(cID);
		fprintf(m_fSceneFile, "\t%sarray<\"%s\">\n", cptrToken, cID);
	}
	else	
		fprintf(m_fSceneFile, "\t%s<%f>\n", cptrToken, key1lst.GetKey(0));
}

void CExport::WriteAsciiKey3Lst(const char cptrToken[], const CInternKey3List &key3lst)
{
	if(key3lst.GetNumKeys()>1)
	{
		char cID[ID_STRING_LENGTH];
		IDtoString(cID);
		fprintf(m_fSceneFile, "\t%sarray<\"%s\">\n", cptrToken, cID);
	}
	else	
		fprintf(m_fSceneFile, "\t%s<%f, %f, %f>\n", cptrToken, key3lst.GetKey(0).x, key3lst.GetKey(0).y, key3lst.GetKey(0).z);
}


//////////////////////////////////////////////////////////////////////////////////

int32 CExport::GenStrips(CInternSurface &surf, unsigned int * pTriIn, int * pTypes,
								int * pLens, unsigned int * pStripIndices, const int32 strip_maxverts)
{
	// verify indata
	assert(strip_maxverts==-1 || strip_maxverts>=3);
	
	// fetch triangle indices
	int iNumTris = 0;
	for(int s=0; s<surf.GetNumTriangles(); s++)
	{
		SInternFace &face = surf.GetFace(s);

		// reject sick triangles
		assert(face.i1!=face.i2 && face.i2!=face.i3 && face.i1!=face.i3);
		
		pTriIn[iNumTris*3+0] = (unsigned int) face.i1;
		pTriIn[iNumTris*3+1] = (unsigned int) face.i2;
		pTriIn[iNumTris*3+2] = (unsigned int) face.i3;
			
		++iNumTris;
	}

	// actc stuff (create strips)
	ACTCData * tc;
	tc = actcNew();
	if(tc == NULL)
	{
		// failed
		exit(0);
	}
	actcParami(tc, ACTC_OUT_MIN_FAN_VERTS, INT_MAX);
	actcParami(tc, ACTC_OUT_MAX_PRIM_VERTS, (strip_maxverts == -1) ? INT_MAX : strip_maxverts);
	actcParami(tc, ACTC_OUT_HONOR_WINDING, ACTC_TRUE);

	int primCount = 0;
	
	primCount = actcTrianglesToPrimitives(tc, iNumTris, (unsigned int (*)[3]) pTriIn,
			pTypes, pLens, pStripIndices, INT_MAX);
	if(primCount < 0)
	{
		exit(0);
	}

	//printf("tris: %d, strips: %d, rejected tris: %d\n", iNumTris, primCount, surf.GetNumTriangles()-iNumTris);
	for(int i=0; i<primCount; i++)
	{
		assert(pTypes[i]==ACTC_PRIM_STRIP);
		//printf("lengths: %d\n", pLens[i]);
	}

	return primCount;
}

// small helper funcs
void CExport::AddInternTextureSeq(CInternTextureSeq * pInternTexSeq)
{
	if(pInternTexSeq != NULL)
		for(int t=0; t<pInternTexSeq->GetNumTexFrames(); t++)	
			AddInternTexture(&pInternTexSeq->GetTexFrame(t));
}

void CExport::AddInternTexture(const CInternTexture * pInternTex)
{
	// look for duplicate
	bool bFound = false;
	int i=0;
	while(i<m_iNumUniqueTextures && !bFound)
		if(*m_pUniqueTextures[i]==*pInternTex)
			bFound = true;
		else
			++i;

	// we alone?
	if(bFound == false)
	{
		// increase array length
		assert(i == m_iNumUniqueTextures);
		const CInternTexture ** pOldArray = m_pUniqueTextures;
		m_pUniqueTextures = NULL;
		m_pUniqueTextures = new const CInternTexture *[m_iNumUniqueTextures+1];
		assert(m_pUniqueTextures != NULL);

		// copy old entries and new entry to new buffer
		for(int i=0; i<m_iNumUniqueTextures; i++)
			m_pUniqueTextures[i] = pOldArray[i];
		m_pUniqueTextures[m_iNumUniqueTextures++] = pInternTex;

		// remove old array
		if(pOldArray != NULL)
			delete [] pOldArray;
	}
}

// helper func for writing texture sequences into material section of scenefile
void CExport::WriteTexSeqToSceneFile(FILE * fptr, const CInternTextureSeq * pTexSeq, const char prefix[], const int32 IDs[])
{
	if(pTexSeq != NULL)
	{
		fprintf(fptr, "\t%s<", prefix);

		for(int f=0; f<pTexSeq->GetNumTexFrames(); f++)
		{
			const int32 index = LocatePosition( &pTexSeq->GetTexFrame(f) );

			const int tmp = IDs[index];
			const char c1 = (tmp>>0)&0xff;
			const char c2 = (tmp>>8)&0xff;
			const char c3 = (tmp>>16)&0xff;
			const char c4 = (tmp>>24)&0xff;

			fprintf(fptr, "\"%c%c%c%c\"", c1, c2, c3, c4);
			if(f < pTexSeq->GetNumTexFrames()-1)
				fprintf(fptr, ", ");
		}

		if(pTexSeq->GetNumTexFrames()>1)
			fprintf(fptr, ", %f", pTexSeq->GetDuration());
		fprintf(fptr, ">\n");
	}
}

const int32 CExport::LocatePosition( const CInternTexture * pTex)
{
	// lookup position
	int32 i=0;
	bool bFound = false;
	while(!bFound)
		if(*m_pUniqueTextures[i] == *pTex)
			bFound = true;
		else
		{
			++i;
			assert(i<m_iNumUniqueTextures);
		}

	return i;
}


/////////////// Bin file utility functions ///////////////

void CExport::PushBinChunk()
{
	int32 iID = MakeNewID();
	const char c1 = (iID>>0)&0xff;
	const char c2 = (iID>>8)&0xff;
	const char c3 = (iID>>16)&0xff;
	const char c4 = (iID>>24)&0xff;
	fwrite(&c1, sizeof(int8), 1, m_fBinFile);	// always LSB of ID first
	fwrite(&c2, sizeof(int8), 1, m_fBinFile);
	fwrite(&c3, sizeof(int8), 1, m_fBinFile);
	fwrite(&c4, sizeof(int8), 1, m_fBinFile);

	m_iRecordBinLocation = ftell(m_fBinFile);
	uint32 uVal = 0;
	fwrite(&uVal, sizeof(uint32), 1, m_fBinFile);	// insert dummy entry for now
}

void CExport::PopBinChunk()
{
	const int32 iCnstCur = ftell(m_fBinFile);
	
	int32 iDeltaOffs = iCnstCur - m_iRecordBinLocation;
	assert(iDeltaOffs>=sizeof(uint32));
	fseek(m_fBinFile, -iDeltaOffs, SEEK_CUR);	// let's go backwards
	iDeltaOffs -= sizeof(uint32);
	uint32 uVal = (uint32) iDeltaOffs;
	fwrite(&uVal, sizeof(uint32), 1, m_fBinFile);	// insert actual size of chunk
	fseek(m_fBinFile, iDeltaOffs, SEEK_CUR);	// let's return

	assert(iCnstCur == ftell(m_fBinFile));
}

int32 CExport::MakeNewID()
{
	int r;
	char id_string[4];
	bool bValid;

	do
	{
		for(int k=0; k<4; k++)
		{
			do
			{
				id_string[k]=rand()&0xff;
			}
			while(!isalpha(id_string[k]) && !isdigit(id_string[k]));
		}
		
		const int r1 = ((int) id_string[0])&0xff;
		const int r2 = ((int) id_string[1])&0xff;
		const int r3 = ((int) id_string[2])&0xff;
		const int r4 = ((int) id_string[3])&0xff;
		
		r = (r1<<0)|(r2<<8)|(r3<<16)|(r4<<24);

		int i=0;
		while(i<m_iTotIDs && r!=m_pTotIDs[i])
			++i;
		
		bValid = i==m_iTotIDs;	// valid if ID was unique
	}
	while(!bValid);

	m_pTotIDs[m_iIndexCounter++] = r;
	return r;
}

void CExport::IDtoString(char * cptrReceiver)
{
	const int tmp = m_pTotIDs[m_iIndexCounter++];
	const char c1 = (tmp>>0)&0xff;
	const char c2 = (tmp>>8)&0xff;
	const char c3 = (tmp>>16)&0xff;
	const char c4 = (tmp>>24)&0xff;

	cptrReceiver[0] = c1;
	cptrReceiver[1] = c2;
	cptrReceiver[2] = c3;
	cptrReceiver[3] = c4;
	cptrReceiver[4] = '\0';
}


