#ifndef __EXPORT_H__
#define __EXPORT_H__

#include "exttypes.h"
#include <stdio.h>

#define ID_STRING_LENGTH (4+1)		// 1 extra for \0 termination


class CInternScene;
class CInternSurface;
class CTransformation;
class CInternTextureSeq;
class CInternTexture;
class CInternKey1List;
class CInternKey3List;
class CInternQuatList;


class CExport
{
public:
	void WriteScene(CInternScene &sceneIn, const char outFolder[], const int32 strip_maxverts);
	
	CExport() : m_pUniqueTextures(NULL), m_iNumUniqueTextures(0), m_fBinFile(NULL), m_pTotIDs(NULL), m_fSceneFile(NULL) {}

private:
	int32 GenStrips(CInternSurface &surf, unsigned int *, int *, int *, unsigned int *, const int32 strip_maxverts);

private:
	void AddInternTextureSeq(CInternTextureSeq * pInternTexSeq);
	void AddInternTexture(const CInternTexture * pInternTex);

	const CInternTexture ** m_pUniqueTextures;

private:
	void WriteTexSeqToSceneFile(FILE * fptr, const CInternTextureSeq * pTexSeq, const char prefix[], const int32 IDs[]);
	const int32 LocatePosition( const CInternTexture * pTex);

private:	// writers to binary file
	void WriteGeometryToBinFile(CInternScene &sceneIn, const int32 strip_maxverts);
	void WriteTexturesToBinFile(CInternScene &sceneIn);
	void WriteSplinesToBinFile(CInternScene &sceneIn);

	void WriteBinKey1List(const CInternKey1List &keyList);
	void WriteBinKey3List(const CInternKey3List &keyList);
	void WriteBinQuatList(const CInternQuatList &keyList);

	int32 m_iMaxTriCount, m_iIndexCounter;
	int32 m_iTotNumSurfs, m_iNumSplines, m_iNumUniqueTextures;
	int32 m_iTotIDs, * m_pTotIDs;
	FILE * m_fBinFile;

private:	// writers to ascii scene file
	void WriteSceneToAscFile(CInternScene &sceneIn);
	void WriteAsciiTransformation(const CTransformation &trans);
	void WriteAsciiKey1Lst(const char cptrToken[], const CInternKey1List &key1lst);
	void WriteAsciiKey3Lst(const char cptrToken[], const CInternKey3List &key3lst);
	
	FILE * m_fSceneFile;

private:	// utility functions
	void PushBinChunk();
	void PopBinChunk();
	int32 MakeNewID();
	void IDtoString(char * cptrReceiver);

	int32 m_iRecordBinLocation;
};


#endif
