#ifndef __INTERNSURFACE_H__
#define __INTERNSURFACE_H__

#include <string>
#include "geommath.h"


class CInternTextureSeq;


class CInternMaterial
{
	friend class CConv3DS;

public:
	const char * GetName() const { return m_matName; }

	CInternTextureSeq * GetDiffSequence() { return m_pDiffTextureSeq; }
	CInternTextureSeq * GetSpecSequence() { return m_pSpecTextureSeq; }
	CInternTextureSeq * GetReflSequence() { return m_pReflTextureSeq; }
	CInternTextureSeq * GetBumpSequence() { return m_pBumpTextureSeq; }

	CInternMaterial() : m_pDiffTextureSeq(NULL), m_pSpecTextureSeq(NULL), m_pReflTextureSeq(NULL), m_pBumpTextureSeq(NULL) {}
	
private:
	char * m_matName;

	CInternTextureSeq * m_pDiffTextureSeq;
	CInternTextureSeq * m_pSpecTextureSeq;
	CInternTextureSeq * m_pReflTextureSeq;
	CInternTextureSeq * m_pBumpTextureSeq;
};

struct SInternFace
{
	SInternFace() {}
	SInternFace(int32 a, int32 b, int32 c) : i1(a), i2(b), i3(c) {}
	
	int32 i1, i2, i3;
};

class CInternSurface
{
	friend class CConv3DS;

public:
	CInternMaterial * GetMaterial() { return m_pMaterial; }
	int32 GetNumTriangles()	{ return m_iNumFaces; }
	int32 GetArrayLen()	{ return m_iLenArrays; }
	Vec3& GetVertex(int32 iVertNum) { return m_pVerts[iVertNum]; }
	Vec3& GetNormal(int32 iNormalNum) { return m_pNormals[iNormalNum]; }
	Vec3& GetTangent(int32 iTangentNum) { return m_pTangents[iTangentNum]; }
	Vec3& GetBiNorm(int32 iBiNormNum) { return m_pBiNorms[iBiNormNum]; }
	Vec3& GetST(int32 iSTNum) { return m_pSTs[iSTNum]; }
	SInternFace& GetFace(int32 iFaceNum) { return m_pFaces[iFaceNum]; }
	
private:
	CInternMaterial * m_pMaterial;
	
	int32 m_iNumFaces;
	SInternFace * m_pFaces;

	int32 m_iLenArrays;
	Vec3 * m_pVerts;
	Vec3 * m_pNormals;
	Vec3 * m_pTangents;
	Vec3 * m_pBiNorms;
	Vec3 * m_pSTs;
};


#endif
