/************************************
 * 3ds C++ File Loader					*
 * Author: Andrea Ingegneri			*
 * modifications: Morten Mikkelsen  *
 * Date: 16/12/2001						*
 * Distribution: LGPL (www.gnu.org) *
 ************************************/

#include <iostream>
#include <string>
#include <fstream>
#include <vector>
#include <cmath>


class C3dsFileFormat
{
public:
	C3dsFileFormat() {}
	~C3dsFileFormat() {}
protected:
#pragma pack(2)
	typedef struct _chunk3ds
	{
		unsigned short	Flag;
		unsigned int	Size;
	} chunk3ds;
#pragma pack()

typedef struct _Transform3dsMatrix
{
	float _11, _12, _13;
	float _21, _22, _23;
	float _31, _32, _33;
} Transform3dsMatrix;

typedef struct _Translate3dsMatrix
{
	float _11, _12, _13;
} Translate3dsMatrix;

class ViewPortLayout
{
public:
	ViewPortLayout(unsigned short Style, short Active, short Unknow1, short Swap, 
				   short Unknow2, short Swap_prior, short Swap_view) : style(Style),
																	   active(Active),
																	   unknow1(Unknow1),
																	   swap(Swap),
																	   unknow2(Unknow2),
																	   swap_prior(Swap_prior),
																	   swap_view(Swap_view) {}
	unsigned short	style;
    short			active;
	short			unknow1;
    short			swap;
	short			unknow2;
    short			swap_prior;
    short			swap_view;
};

enum {CameraNameSize = 11};

class ViewPortData
{
public:
	ViewPortData(short Flags, short Axis_lockout, short Win_x, short Win_y, short Win_w, short Win_h,
				 short Win_view, float Zoom, float Worldcenter_x, float Worldcenter_y, float Worldcenter_z,
				 float Horiz_ang, float Vert_ang, char *CameraName) : flags(Flags), axis_lockout(Axis_lockout),
																	  win_x(Win_x), win_y(Win_y),
																	  win_w(Win_w), win_h(Win_h),
																	  win_view(Win_view), zoom(Zoom),
																	  worldcenter_x(Worldcenter_x),
																	  worldcenter_y(Worldcenter_y),
																	  worldcenter_z(Worldcenter_z),
																	  horiz_ang(Horiz_ang), vert_ang(Vert_ang),
																	  camera_name(CameraName) {}
	short	flags, axis_lockout;
	short	win_x, win_y, win_w, win_h, win_view;
	float	zoom; 
	float	worldcenter_x, worldcenter_y, worldcenter_z;
	float	horiz_ang, vert_ang;
	string	camera_name;
};

enum {
	CHUNK_VERSION	= 0x0002,
    CHUNK_RGBF      = 0x0010,
    CHUNK_RGBB      = 0x0011,

	CHUNK_PERCENTW	= 0x0030,
	CHUNK_PERCENTF	= 0x0031,

    CHUNK_PRJ       = 0xC23D,
    CHUNK_MLI       = 0x3DAA,

    CHUNK_MAIN      = 0x4D4D,
        CHUNK_OBJMESH   = 0x3D3D,
			CHUNK_ONEUNIT	= 0x0100,
            CHUNK_BKGCOLOR  = 0x1200,
            CHUNK_AMBCOLOR  = 0x2100,
			CHUNK_DEFAULT_VIEW = 0x3000,
				CHUNK_VIEW_TOP = 0x3010,
				CHUNK_VIEW_BOTTOM = 0x3020,
				CHUNK_VIEW_LEFT = 0x3030,
				CHUNK_VIEW_RIGHT = 0x3040,
				CHUNK_VIEW_FRONT = 0x3050,
				CHUNK_VIEW_BACK = 0x3060,
				CHUNK_VIEW_USER = 0x3070,
				CHUNK_VIEW_CAMERA = 0x3080,
            CHUNK_OBJBLOCK  = 0x4000,
                CHUNK_TRIMESH   = 0x4100,
                    CHUNK_VERTLIST  = 0x4110,
                    CHUNK_VERTFLAGS = 0x4111,
                    CHUNK_FACELIST  = 0x4120,
                    CHUNK_FACEMAT   = 0x4130,
                    CHUNK_MAPLIST   = 0x4140,
                    CHUNK_SMOOLIST  = 0x4150,
                    CHUNK_TRMATRIX  = 0x4160,
                    CHUNK_MESHCOLOR = 0x4165,
                    CHUNK_TXTINFO   = 0x4170,
                CHUNK_LIGHT     = 0x4600,
                    CHUNK_SPOTLIGHT = 0x4610,
						  CHUNK_SPOT_RECTANGULAR = 0x4651,
						  CHUNK_SPOT_ROLL	  = 0x4656,
						  CHUNK_SPOT_ASPECT = 0x4657,
						  CHUNK_INNER_RANGE = 0x4659,
						  CHUNK_OUTER_RANGE = 0x465a,
                CHUNK_CAMERA    = 0x4700,
                CHUNK_HIERARCHY = 0x4F00,

        CHUNK_VIEWPORT_LAYOUT_OLD	= 0x7000,
			CHUNK_VIEWPORT_DATA_OLD	= 0x7010,
				CHUNK_VIEWPORT_SIZE = 0x7020,
					CHUNK_NETWORK_VIEW = 0X7030,

        CHUNK_VIEWPORT_LAYOUT	= 0x7001,
			CHUNK_VIEWPORT_DATA	= 0x7011,
			CHUNK_VIEWPORT_DATA3 = 0x7012,

		CHUNK_MATERIAL  = 0xAFFF,
            CHUNK_MATNAME   	= 0xA000,

				// colors
            CHUNK_AMBIENT   	= 0xA010,
            CHUNK_DIFFUSE   	= 0xA020,
            CHUNK_SPECULAR  	= 0xA030,

				// colors by mapfile
            CHUNK_DIFFUSE_MAP   	= 0xA200,
				CHUNK_SPECULAR_MAP	= 0xA204,
				CHUNK_REFLECTION_MAP	= 0xA220,
            CHUNK_BUMPMAP_MAP  	= 0xA230,
        		    CHUNK_MAPFILE   = 0xA300,
					 
        CHUNK_KEYFRAMER = 0xB000,
            CHUNK_AMBIENTKEY    = 0xB001,
            CHUNK_TRACKINFO = 0xB002,
               CHUNK_TRACKOBJNAME  = 0xB010,
			   	CHUNK_TRACKINSTNAME = 0xB011,
               CHUNK_TRACKPIVOT    = 0xB013,
               CHUNK_TRACKPOS      = 0xB020,
               CHUNK_TRACKROTATE   = 0xB021,
               CHUNK_TRACKSCALE    = 0xB022,
               CHUNK_TRACKMORPH    = 0xB026,
               CHUNK_TRACKHIDE     = 0xB029,
               CHUNK_OBJNUMBER     = 0xB030,
            CHUNK_TRACKCAMERA = 0xB003,
					CHUNK_TRACKFOV  = 0xB023,
					CHUNK_TRACKROLL = 0xB024,
            CHUNK_TRACKCAMTGT = 0xB004,
            CHUNK_TRACKLIGHT  = 0xB005,
            CHUNK_TRACKLIGTTGT = 0xB006,
            CHUNK_TRACKSPOTL  = 0xB007,
					CHUNK_TRACKHOTSPOT	= 0xB027,
					CHUNK_TRACKFALLOFF	= 0xB028,
            CHUNK_FRAMES    = 0xB008,
	};
};

class C3dsFileLoader : public C3dsFileFormat
{
public:
// Constructor
	C3dsFileLoader() : C3dsFileFormat(), PC(0), mFileSize(0), m_iCurDepth(0) { m_eModeStack[m_iCurDepth]=UNSPECIFIED_PARSE; }
	virtual ~C3dsFileLoader() {}

// Methods
	void		ProcessFile(string FileName);
private:
	void		movePC(unsigned int forward);
	bool		ParseChunks();
	bool		ParseSubChunks(int endPos, const int level);
	bool		ReadBlock(char *dest, unsigned int size);
	float		ReadRealNumber();
	bool		SkipChunk(chunk3ds &chunk);
	void		FileSize();
	bool		isOk();
	void		ReadASCIIZ(char *dest);
	void		ReadRGBF();
	void		ReadRGBB();
	void		ReadObjBlock();
	void		ReadVertList();
	void		ReadFaceList();
	void		ReadFaceMat();
	void 		ReadSmoothList(const unsigned int size_chnk_data);
	void		ReadMapList();
	void		ReadTrMatrix();
	void		ReadLight();
	void		ReadSpotLight();
	void		ReadSpotRectangular();
	void		ReadSpotRoll();
	void		ReadSpotAspect();
	void		ReadInnerRange();
	void		ReadOuterRange();
	void		ReadTriMesh();
	void		ReadCamera();
	void		ReadMaterial();
	void		ReadMatName();
	void		ReadMapFile();
	void		ReadFrames();
	void		ReadTrackObjName();
	void		ReadTrackInstName();
	void		ReadPivotPoint();
	void		ReadTrack(unsigned short Flag);
	void		ReadObjNumber();
	void		ReadObjChunk();
	void		ReadOneUnit();
	void		ReadViewPortData();
	void		ReadViewUser();
	void		ReadViewTop();
	void		ReadViewBottom();
	void		ReadViewLeft();
	void		ReadViewRight();
	void		ReadViewFront();
	void		ReadViewBack();
	void		ReadViewCamera();
	void		ReadViewPortLayout();
	void		ReadViewPortSize();
	void		EndReading();
	
	ifstream	mInputFile;
protected:
	unsigned int		PC;		// File Pointer
	unsigned int		mFileSize;
// Inheritable functions for polymorphism
	virtual void		UserRGBF(float r, float g, float b);
	virtual void		User3dVert(float x, float y, float z);
	virtual void		User3dFace(unsigned short A, unsigned short B, unsigned short C, unsigned short Flags);
	virtual void		UserTriMesh();
	virtual void		UserCamera(float px, float py, float pz,
								   float tx, float ty, float tz,
								   float Bank, float Lens);	
	virtual void		UserMaterial();
	virtual void		UserFaceMaterial(string Name, int Number);
	virtual void		UserFaceSmoothGroup(unsigned int smoothRef);
	virtual void		UserMapVertex(float U, float V);
	virtual void		UserTransformMatrix(const Transform3dsMatrix &Transform, const Translate3dsMatrix &Translate);
	virtual void		UserLight(float x, float y, float z);
	virtual void		UserSpotLight(float x, float y, float z, float HotSpot, float FallOff);
	virtual void		UserSetSpotRectangular();
	virtual void		UserSpotRoll(float roll);
	virtual void		UserSpotAspect(float aspect);
	virtual void		UserInnerRange(float inner);
	virtual void		UserOuterRange(float outer);
	virtual void		UserMatName(const string Name);
	virtual void		UserMapFile(const string FileName);
	virtual void		UserFrames(long Start, long End);
	virtual void		UserTrackInfo() {}
	virtual void		UserTrackCamera() {}
	virtual void		UserTrackCamTgt() {}
	virtual void		UserTrackLight() {}
	virtual void		UserTrackLightTgt() {}
	virtual void		UserTrackSpotL() {}
	virtual void		UserTrackObjectName(const string Name, int Key1, int Key2, int Key3);
	virtual void		UserTrackInstName( const string trackInstName );
	virtual void		UserPivotPoint(float PivX, float PivY, float PivZ);
	virtual void		UserTrackPos(int Frame, float t, float c, float b, float et, float ef, float x, float y, float z);
	virtual void		UserTrackRot(int Frame, float t, float c, float b, float et, float ef, float x, float y, float z, float w);
	virtual void		UserTrackScale(int Frame, float t, float c, float b, float et, float ef, float ScaleX, float ScaleY, float ScaleZ);
	virtual void		UserTrackFov(int Frame, float t, float c, float b, float et, float ef, float fov);
	virtual void		UserTrackRoll(int Frame, float t, float c, float b, float et, float ef, float roll);
	virtual void		UserTrackHotSpot(int Frame, float t, float c, float b, float et, float ef, float hotspot);
	virtual void		UserTrackFallOff(int Frame, float t, float c, float b, float et, float ef, float falloff);
	virtual void		UserObjNumber(int ObjNumber);
	virtual void		UserObjName(string Name);
	virtual void		UserChunkObj();
	virtual void		UserOneUnit(float Unit);
	virtual void		UserViewPortLayout(ViewPortLayout &Layout);
	virtual void		UserViewPortSize(int PosX, int PosY, int Width, int Height);
	virtual void		UserViewPortData(ViewPortData &Data);
	virtual	void		UserViewUser(float x, float y, float z, float Width, float XYangle, float YZangle, float BackAngle);
	virtual void		UserViewCamera(string CameraName);
	virtual void		UserEndOfFile();

	// Identifying current subject of parsing
protected:
	enum PARSE_MODE	// push stack modes
	{
		UNSPECIFIED_PARSE=0,
		DIFFUSE_PARSE,
		SPECULAR_PARSE,
		BUMPMAP_PARSE,
		REFLECTION_PARSE,
		LIGHT_PARSE,

		TRACKINFO_PARSE,
		TRACKCAMERA_PARSE,
		TRACKCAMTGT_PARSE,
		TRACKLIGHT_PARSE,
		TRACKLIGHTTGT_PARSE,
		TRACKSPOTL_PARSE,
	
		// end
		NUM_PARSE_MODES
	};
	
	PARSE_MODE GetCurMode() { return m_eModeStack[m_iCurDepth]; }

private:	
	void PushMode(PARSE_MODE mode) { m_eModeStack[++m_iCurDepth] = mode; }
	void PopMode() { --m_iCurDepth; }

	int m_iCurDepth;
	PARSE_MODE m_eModeStack[NUM_PARSE_MODES];
};
